import pickle
from typing_extensions import Any, Callable, List, Optional

import numpy as np

from escape.core.array import double_array
from escape.core.model import modelstack_obj
from escape.core.regressor import regressor_obj


class BaseRegressor(regressor_obj):
    """Base class for regressors.
    This base class is derived from regressor_obj and is used to define the
    common interface for all regressors. It is not meant to be used directly
    but to be subclassed by other regressors.
    """

    __version__: int = 1

    def __init__(
        self,
        name: str,
        mstack: modelstack_obj,
        *,
        nsamples: Any = 1000,
        ntests: Any = 100,
        data_noisifiers: Optional[List[Callable]] = None,
        coordinates_noisifiers: Optional[List[Callable]] = None,
        data_normalizers: Optional[List[Callable]] = None
    ):
        """
        Initializes the regressor. Data noisifiers, coordinates noisifiers and data normalizers
        are a list of callables. This callables are used to noisify the data, noisify the coordinates
        and normalize the data, respectively. Noisifying and normalization of the training data is
        important to avoid overfitting and to improve the generalization of the model.

        :Parameters:

        name: str
            Name of the regressor.
        mstack: modelstack_obj
            Model stack.
        nsamples: Any
            Number of training samples.
        ntests: Any
            Number of testing samples.
        data_noisifiers: Optional[List[Callable]] - List of callables
            Data noisifiers.
        coordinates_noisifier: Optional[Callable] - List of callables
            Coordinates noisifiers.
        data_normalizers: Optional[List[Callable]] - List of callables
            Data normalizers.
        """

        regressor_obj.__init__(self, name, mstack, nsamples=nsamples, ntests=ntests)

        data_noisifiers = data_noisifiers if data_noisifiers is not None else []
        coordinates_noisifiers = (
            coordinates_noisifiers if coordinates_noisifiers is not None else []
        )
        data_normalizers = data_normalizers if data_normalizers is not None else []

        if not isinstance(data_noisifiers, (list, tuple)):
            data_noisifiers = [data_noisifiers]
        if not isinstance(coordinates_noisifiers, (list, tuple)):
            coordinates_noisifiers = [coordinates_noisifiers]
        if not isinstance(data_normalizers, (list, tuple)):
            data_normalizers = [data_normalizers]
        if not all(
            [
                callable(it)
                for it in data_noisifiers + data_normalizers + coordinates_noisifiers
            ]
        ):
            raise ValueError("Data transformers expected to be callable objects")

        self._data_normalizers = data_normalizers
        self._data_noisifiers = data_noisifiers
        self._coordinates_noisifiers = coordinates_noisifiers

    def normalize_data(
        self, X: np.ndarray, c: np.ndarray, reset: bool = True
    ) -> np.ndarray:
        """
        Normalizes the training data. This method calls the data normalizers in the order they were
        added to the regressor.

        :Parameters:

        X: np.ndarray
            Training data.
        c: np.ndarray
            Coordinates.
        reset: bool
            Reset the normalizers.

        :Returns:

        np.ndarray
            Normalized data.
        """

        methods = self._data_normalizers
        if not methods:
            return X
        for m in methods:
            res = m(X, c, reset=reset)
        return res

    def noisify_data(self, X: np.ndarray, reset: bool = True) -> np.ndarray:
        """
        Noisifies the training data. This method calls the data noisifiers in the order they were
        added to the regressor.

        :Parameters:

        X: np.ndarray
            Training data.
        reset: bool
            Reset the noisifiers.

        :Returns:

        np.ndarray
            Noisified data.
        """
        methods = self._data_noisifiers
        if not methods:
            return X
        for m in methods:
            res = m(X, reset=reset)
        return res

    def noisify_coords(self, c: np.ndarray, reset: np.ndarray = True) -> np.ndarray:
        """
        Noisifies the coordinates. This method calls the coordinates noisifiers in the order they were
        added to the regressor.

        :Parameters:

        c: np.ndarray
            Coordinates.

        reset: bool
            Reset the noisifiers.

        :Returns:

        np.ndarray
            Noisified coordinates.
        """
        methods = self._coordinates_noisifiers
        if not methods:
            return c
        res = np.asarray(c)
        for m in methods:
            res = m(res, reset=reset)
        res = np.asarray(res)
        return res

    def dumps(self) -> bytes:
        """
        Dumps the regressor attributes as a string.
        This method is automatically called by the serialization mechanism.
        It can be overridden by subclasses to serialize the regressor attributes.

        :Returns:

        bytes
            Serialized regressor attributes.
        """
        # dump data noisifiers, coordinates noisifiers and data normalizers
        # as a string using pickle
        transformers = [
            self._data_noisifiers,
            self._coordinates_noisifiers,
            self._data_normalizers,
        ]
        out = pickle.dumps(
            (self.__version__, transformers), protocol=pickle.HIGHEST_PROTOCOL
        )
        return out

    def loads(self, inp: bytes) -> None:
        """
        Loads the regressor attributes from a tuple of strings.
        This method is automatically called by the serialization mechanism.
        It can be overridden by subclasses to deserialize the regressor attributes.

        :Parameters:
        bytes: tuple
            Tuple containing the serialized regressor attributes.
        """
        try:
            version, data = pickle.loads(inp)
        except Exception as e:
            raise ValueError("Invalid input") from e
        if version == 1:
            # load data noisifiers, coordinates noisifiers and data normalizers
            # from a string using pickle
            if not isinstance(data, (list, tuple)):
                raise ValueError("Invalid input, expected list or tuple")
            if len(data) != 3:
                raise ValueError("Invalid input, expected 3 elements")
            self._data_noisifiers = data[0]
            self._coordinates_noisifiers = data[1]
            self._data_normalizers = data[2]
        else:
            raise ValueError("Unsupported version")
