import io
import logging
import pickle
import traceback
from typing_extensions import Any, Callable, List, Optional, Type, Union

import numpy as np
import skorch
import torch
from escape.core.setting import int_setting_obj
from escape.utils.regressor.base import BaseRegressor
from typeguard import typechecked

from escape.core.model import modelstack_obj

from .mlp import MLP
from .module import Module

logger = logging.getLogger(__name__)


class Regressor(skorch.NeuralNetRegressor, BaseRegressor):
    """Pytorch regressor using skorch and ESCAPE regressor interface.

    Args:
        module (torch.nn.Module): PyTorch module

    Keyword Args (skorch.NeuralNetRegressor):
        criterion (torch.nn.Module): PyTorch module
        optimizer (torch.optim.Optimizer): PyTorch optimizer
        lr (float): Learning rate
        max_epochs (int): Maximum number of epochs
        batch_size (int): Batch size
        device (str): Device to use for training
        train_split (None or callable): If None, use the training data. If callable, must return a tuple of (train dataset, validation dataset).
        callbacks (list): List of callbacks
        warm_start (bool): If True, continue training from the last state
        verbose (int): Verbosity level
    """

    # version of the regressor for serialization
    __version__: int = 1

    @typechecked
    def __init__(
        self,
        name: str,
        mstack: modelstack_obj,
        *,
        nsamples: Union[int, int_setting_obj],
        ntests: Union[int, int_setting_obj],
        module: Union[Type[Module], Module] = MLP,
        data_noisifiers: Optional[List[Callable]] = None,
        coordinates_noisifiers: Optional[List[Callable]] = None,
        data_normalizers: Optional[List[Callable]] = None,
        **kwargs,
    ):
        """
        Initializes the regressor. Data noisifiers, coordinates noisifiers and data normalizers
        are a list of callables. This callables are used to noisify the data, noisify the coordinates
        and normalize the data, respectively. Noisifying and normalization of the training data is
        important to avoid overfitting and to improve the generalization of the model.

        :Parameters:

        name: str
            Name of the regressor.
        mstack: modelstack_obj
            Model stack.
        nsamples: int
            Number of training samples.
        ntests: int
            Number of testing samples.
        data_noisifiers: Optional[List[Callable]] - List of callables
            Data noisifiers.

        coordinates_noisifier: Optional[Callable] - List of callables
            Coordinates noisifiers.
        data_normalizers: Optional[List[Callable]] - List of callables
            Data normalizers.

        module: Module or type
            Module or subclass of Module

        kwargs: dict
            Keyword arguments for skorch.NeuralNetRegressor
        """
        BaseRegressor.__init__(
            self,
            name,
            mstack,
            nsamples=nsamples,
            ntests=ntests,
            data_noisifiers=data_noisifiers,
            coordinates_noisifiers=coordinates_noisifiers,
            data_normalizers=data_normalizers,
        )
        if isinstance(module, type) and issubclass(module, Module):
            ntrp = self.num_train_params
            logger.info(f"Number of training parameters: {ntrp}")
            if ntrp == 0:
                raise ValueError("Number of training parameters is zero")
            dl = self.data_length
            logger.info(f"Data length: {dl}")
            module = module(dl, ntrp)
            logger.info(f"Module: {module}")
        elif isinstance(module, Module):
            pass
        skorch.NeuralNetRegressor.__init__(self, module=module, **kwargs)

    def fit(self, X: np.ndarray, y: np.ndarray = None, **fit_params):
        """Fit the model to the training data.

        :Parameters:

        X: np.ndarray
            Training data.
        y: np.ndarray
            Target values.

        :Returns:

        self: Regressor
            Returns self.
        """
        if self.is_stopped:
            logger.info("Fit stopped")
            return None
        # convert to float32 to avoid errors in skorch
        X = np.asarray(X, dtype=np.float32)
        y = np.asarray(y, dtype=np.float32)
        logger.info(f"X: {X.shape}, y: {y.shape}")
        res = super().fit(X, y, **fit_params)
        return res

    def predict(self, X):
        """Predicts the target values.

        :Parameters:

        X: np.ndarray
            Input data.

        :Returns:

        np.ndarray
            Predicted target values.
        """
        if self.is_stopped:
            logger.info("Predict stopped")
            return None
        # convert to float32 to avoid errors in skorch
        X = np.asarray(X, dtype=np.float32)
        onedim = False
        if X.ndim == 1:
            onedim = True
            # If the input is a single data array, add a batch dimension
            X = X.reshape(1, -1)

        res = super().predict(X)
        logger.info(f"Res: {res.shape}")
        if onedim:
            res = res[0]
        return res

    def score(self, X, y):
        """Returns the score of the model.

        :Parameters:

        X: np.ndarray
            Input data.
        y: np.ndarray
            Target values.

        :Returns:

        float
            Score.
        """
        # convert to float32 to avoid errors in skorch
        X = np.asarray(X, dtype=np.float32)
        y = np.asarray(y, dtype=np.float32)
        res = super().score(X, y)
        return res

    def dumps(self) -> bytes:
        sbase = BaseRegressor.dumps(self)
        try:
            buffer = io.BytesIO()
            m = torch.jit.script(self.module_)
            torch.jit.save(
                m,
                buffer,
            )
            smdl = buffer.getvalue()
        except Exception as e:
            traceback.print_exc()
            raise ValueError(f"Error while serializing the model: {e}")
        try:
            out = pickle.dumps((self.__version__, sbase, smdl))
        except Exception as e:
            traceback.print_exc()
            raise ValueError(f"Error while serializing the regressor: {e}")
        return out

    def loads(self, inp: bytes) -> None:
        version, sbase, smdl = pickle.loads(inp)
        BaseRegressor.loads(self, sbase)
        if version == 1:
            buffer = io.BytesIO(smdl)
            try:
                mdl = torch.jit.load(buffer, map_location=torch.device("cpu"))
                # evaluating the model to initialize the parameters
                mdl.eval()
                skorch.NeuralNetRegressor.__init__(self, module=mdl)
                self.initialize()
            except Exception as e:
                traceback.print_exc()
                raise ValueError(f"Invalid model input format: {e}")
            
    def run_single_epoch(self, iterator, training, prefix, step_fn, **fit_params):
        """Compute a single epoch of train or validation.

        Parameters
        ----------
        iterator : torch DataLoader or None
            The initialized ``DataLoader`` to loop over. If None, skip this step.
        training : bool
            Whether to set the module to train mode or not.
        prefix : str
            Prefix to use when saving to the history.
        step_fn : callable
            Function to call for each batch.
        **fit_params : dict
            Additional parameters passed to the ``step_fn``.
        """
        if self.is_stopped:
            raise RuntimeError("Training stopped")
        return super().run_single_epoch(iterator, training, prefix, step_fn, **fit_params)
        
        
        


from escape.serialization import dumps_regressor, loads_regressor


def loads_torch_regressor(s: bytes) -> Regressor:
    obj = Regressor.__new__(Regressor)
    return loads_regressor(s, obj)


def reduce_torch_regressor(obj):
    return (loads_torch_regressor, (dumps_regressor(obj),))


import copyreg

copyreg.pickle(Regressor, reduce_torch_regressor)
