from typing_extensions import Any

import torch
import torch.nn as nn

from .module import Module


class IdenttityBlock(nn.Module):

    def __init__(
        self, in_dim: int, out_dim: int, activation: Any = nn.ReLU(), bias: bool = True
    ):
        super(IdenttityBlock, self).__init__()
        self.activation = activation
        self.lin1 = nn.Linear(in_dim, out_dim, bias=bias)
        self.bn1 = nn.BatchNorm1d(out_dim)
        self.lin2 = nn.Linear(out_dim, out_dim, bias=bias)
        self.bn2 = nn.BatchNorm1d(out_dim)
        self.lin3 = nn.Linear(out_dim, out_dim, bias=bias)
        self.bn3 = nn.BatchNorm1d(out_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.lin1(x)
        out = self.bn1(out)
        out = self.activation(out)
        out = self.lin2(out)
        out = self.bn2(out)
        out = self.activation(out)
        out = self.lin3(out)
        out = self.bn3(out)
        out += x
        out = self.activation(out)
        return out


class DenseBlock(nn.Module):

    def __init__(
        self, in_dim: int, out_dim: int, activation: Any = nn.ReLU(), bias: bool = True
    ):
        super(DenseBlock, self).__init__()
        self.activation = activation
        self.lin1 = nn.Linear(in_dim, out_dim, bias=bias)
        self.bn1 = nn.BatchNorm1d(out_dim)
        self.lin2 = nn.Linear(out_dim, out_dim, bias=bias)
        self.bn2 = nn.BatchNorm1d(out_dim)
        self.lin3 = nn.Linear(out_dim, out_dim, bias=bias)
        self.bn3 = nn.BatchNorm1d(out_dim)
        self.shortcut = nn.Linear(in_dim, out_dim, bias=bias)
        self.shortcutbn = nn.BatchNorm1d(out_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.lin1(x)
        out = self.bn1(out)
        out = self.activation(out)
        out = self.lin2(out)
        out = self.bn2(out)
        out = self.activation(out)
        out = self.lin3(out)
        out = self.bn3(out)
        shortcut = self.shortcut(x)
        shortcut = self.shortcutbn(shortcut)
        out += shortcut
        out = self.activation(out)
        return out


class ResNet(Module):
    def __init__(self, in_dim: int, out_dim: int):
        super(ResNet, self).__init__(in_dim, out_dim)
        self.res_input = nn.Linear(in_dim, 1024)

        self.dense1 = DenseBlock(1024, 1024)
        self.iden1 = IdenttityBlock(1024, 1024)
        self.iden2 = IdenttityBlock(1024, 1024)

        self.dense2 = DenseBlock(1024, 1024)
        self.iden3 = IdenttityBlock(1024, 1024)
        self.iden4 = IdenttityBlock(1024, 1024)

        self.dense3 = DenseBlock(1024, 1024)
        self.iden5 = IdenttityBlock(1024, 1024)
        self.iden6 = IdenttityBlock(1024, 1024)

        self.bn_out = nn.BatchNorm1d(1024)
        self.relu_out = nn.ReLU()
        self.lin_out = nn.Linear(1024, out_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.res_input(x)
        out = self.dense1(out)
        out = self.iden1(out)
        out = self.iden2(out)
        out = self.dense2(out)
        out = self.iden3(out)
        out = self.iden4(out)
        out = self.dense3(out)
        out = self.iden5(out)
        out = self.iden6(out)
        out = self.bn_out(out)
        out = self.relu_out(out)
        out = self.lin_out(out)
        return out
