"""

This module contains classes and functions for data transformation.
The data is expected to be in the form of a 2D numpy array with the shape (n_samples, n_features).
The data trasnform instances can be applied either to features or to samples.

The following classes are defined:

- BaseDataTransformer: Base class for data transformers.
- MinMaxScaler: Scales the data to the given range [min, max].
- MaxSamplesNormalizer: Normalizes the data by the maximum value of each sample.
- LogScaler: Scales the data by the logarithm keeping sample or feature range.
- RandomSchifter: Shifts the data by a random value. It is used mainly for the coordinate transformation.
- RandomMultiplier: Multiplies the data by a random value.
- RandomNormalBackground: Adds a random normal background to the data.
- PoissonBackground: Adds a poisson background to the data.
- PorodsMultiplier: Multiplies the data by the Porod's law.

"""

import sys
from typing_extensions import Optional, Tuple

import numpy as np


class BaseDataTransformer:
    """Base class for data transformers."""

    def __init__(self):
        """Initializes the instance."""
        self._fitted = False
        self._scale = None

    def fit(
        self, X: np.ndarray, c: Optional[np.ndarray] = None
    ) -> "BaseDataTransformer":
        """Fits the transformer to the data.

        :Parameters:

        X: numpy.ndarray
            The data to fit the transformer to.

        c: numpy.ndarray
            The coordinates of the data points. Required for some transformers.
        """
        self._fitted = True
        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Transforms the data.

        :Parameters:

        X: numpy.ndarray
            The data to transform.
        """
        if not self._fitted:
            raise ValueError("The transformer is not fitted")
        return X

    def __call__(
        self,
        X: np.ndarray,
        c: Optional[np.ndarray] = None,
        reset: Optional[bool] = True,
    ) -> np.ndarray:
        """Transforms the data.

        :Parameters:

        X: numpy.ndarray
            The data to transform.
        c: numpy.ndarray
            The coordinates of the data points. Required for some transformers.
        reset: bool
            If True, the transformer is fitted to the data before transforming it.

        :Returns:

        numpy.ndarray
            The transformed data.
        """
        if not self._fitted or reset:
            self.fit(X, c)
            self._fitted = True
        res = self.transform(X)
        return res


class MinMaxScaler(BaseDataTransformer):
    """
    Scales the data to the given range [min, max].
    """

    def __init__(self, feature_range: Tuple = (0, 1), clip: bool = False):
        """Initializes the instance.

        :Parameters:

        feature_range: tuple
            The range of the transformed data.

        clip: bool
            If True, the data is clipped to the feature or sample range.
        """
        BaseDataTransformer.__init__(self)
        fmin, fmax = feature_range
        if abs(fmax - fmin) < sys.float_info.epsilon:
            raise ValueError(
                "The range of the transformed data must be greater than zero"
            )
        self._feature_range = feature_range
        self._scales = None
        self._xmin = None
        self._xmax = None
        self._clip = clip

    def fit(self, X: np.ndarray, c: Optional[np.ndarray] = None) -> "MinMaxScaler":
        fmin, fmax = self._feature_range
        scales = None
        xmin = None
        xmax = None
        if X.ndim == 1:
            X = X.reshape(1, -1)
        xmax = X.max(axis=0)
        xmin = X.min(axis=0)
        abs_diff = np.abs(xmax - xmin)
        scales = np.zeros(xmax.size)
        mask = abs_diff > sys.float_info.epsilon
        scales[mask] = (fmax - fmin) / (xmax - xmin)[mask]
        self._scales = scales
        self._xmin = xmin
        self._xmax = xmax
        return super().fit(X, c)

    def transform(self, X) -> np.ndarray:
        if X.ndim == 1:
            X = X.reshape(1, -1)
        fmin, fmax = self._feature_range
        if self._scales is None:
            raise ValueError("The scales are not available")
        xmin, xmax = self._xmin, self._xmax
        xmin = xmin[np.newaxis, :]
        xmax = xmax[np.newaxis, :]
        scales = self._scales[np.newaxis, :]
        res = (X - xmin) * scales + fmin
        if X.ndim == 1:
            res = res.flatten()
        if self._clip:
            res = np.clip(res, fmin, fmax)
        res = super().transform(res)
        return res


class MaxSampleNormalizer(BaseDataTransformer):
    """
    Normalizes the data by the maximum value of each sample.
    """

    def __init__(self):
        BaseDataTransformer.__init__(self)

    def fit(
        self, X: np.ndarray, c: Optional[np.ndarray] = None
    ) -> "MaxSamplesNormalizer":
        return super().fit(X, c)

    def transform(self, X) -> np.ndarray:
        if X.ndim == 1:
            X = X.reshape(1, -1)
        res = X / X.max(axis=1)[:, np.newaxis]
        res = super().transform(res)
        return res


class LogScaler(BaseDataTransformer):
    """
    Scales the data by the logarithm keeping sample or feature range.
    """

    def __init__(self, apply_to_features: bool = False, base: float = 10):
        """Initializes the instance.

        :Parameters:

        apply_to_features: bool
            If True, the logarithm is applied to features. Otherwise, it is applied to samples.
        base: float
            The base of the logarithm.
        """
        BaseDataTransformer.__init__(self)
        self._apply_to_features = apply_to_features
        self._base = base

    def fit(self, X: np.ndarray, c: Optional[np.ndarray] = None) -> "LogScaler":
        return super().fit(X, c)

    def transform(self, X: np.ndarray) -> np.ndarray:
        if X.ndim == 1:
            X = X.reshape(1, -1)
        logx = np.log(X) / np.log(self._base)
        if self._apply_to_features:
            fmin, fmax = X.min(axis=0), X.max(axis=0)
            xmin, xmax = logx.min(axis=0), logx.max(axis=0)
            abs_diff = np.abs(xmax - xmin)
            scales = np.zeros(xmax.size)
            mask = abs_diff > sys.float_info.epsilon
            scales[mask] = ((fmax - fmin) / (xmax - xmin))[mask]
            xmin = xmin[np.newaxis, :]
            xmax = xmax[np.newaxis, :]
            scales = scales[np.newaxis, :]
            fmin = fmin[np.newaxis, :]
            fmax = fmax[np.newaxis, :]
        else:
            fmin, fmax = X.min(axis=1), X.max(axis=1)
            xmin, xmax = logx.min(axis=1), logx.max(axis=1)
            abs_diff = np.abs(xmax - xmin)
            scales = np.zeros(xmax.size)
            mask = abs_diff > sys.float_info.epsilon
            scales[mask] = ((fmax - fmin) / (xmax - xmin))[mask]
            xmin = xmin[:, np.newaxis]
            xmax = xmax[:, np.newaxis]
            scales = scales[:, np.newaxis]
            fmin = fmin[:, np.newaxis]
            fmax = fmax[:, np.newaxis]
        res = (logx - xmin) * scales + fmin
        res = super().transform(res)
        return res


class RandomSchifter(BaseDataTransformer):
    """
    Shifts the data by a random value. It is used mainly for the coordinate transformation.
    """

    def __init__(self, delta_range: Tuple = (-1e-4, 1e-4)):
        """Initializes the instance.

        :Parameters:

        delta_range: tuple
            The range of the random shift.

        """
        BaseDataTransformer.__init__(self)
        self.delta_range = delta_range

    def fit(self, X: np.ndarray, c: Optional[np.ndarray] = None) -> "RandomSchifter":
        return super().fit(X, c)

    def transform(self, X: np.ndarray) -> np.ndarray:
        res = X + np.random.uniform(*self.delta_range, size=X.shape)
        res = super().transform(res)
        return res


class RandomMultiplier(BaseDataTransformer):
    """
    Multiplies the data by a random value.
    """

    def __init__(self, range: Tuple = (0.999, 1.001)):
        """Initializes the instance.

        :Parameters:

        range: tuple
            The range of the random multiplier.

        """
        BaseDataTransformer.__init__(self)
        self.range = range

    def fit(self, X: np.ndarray, c: Optional[np.ndarray] = None) -> "RandomMultiplier":
        return super().fit(X, c)

    def transform(self, X: np.ndarray) -> np.ndarray:
        res = X * np.random.uniform(*self.range, size=X.shape)
        res = super().transform(res)
        return res


class RandomNormalBackground(BaseDataTransformer):
    """
    Adds a random normal background to the data.
    """

    def __init__(
        self, mean_range: Tuple = (1e-6, 1e-4), apply_to_features: bool = False
    ):
        """Initializes the instance.

        :Parameters:

        mean_range: tuple
            The range of the mean of the random normal background.

        apply_to_features: bool
            If True, the background is applied to features. Otherwise, it is applied to samples.
        """
        BaseDataTransformer.__init__(self)
        self._mean_range = mean_range
        self._apply_to_features = apply_to_features

    def fit(
        self, X: np.ndarray, c: Optional[np.ndarray] = None
    ) -> "RandomNormalBackground":
        return super().fit(X, c)

    def transform(self, X: np.ndarray) -> np.ndarray:
        # for features we create a random normal background for each feature(column)
        # and repeat this array for each sample(row)
        mr = self._mean_range
        if self._apply_to_features:
            if X.ndim == 1:
                b = np.random.uniform(*mr, size=X.size)
            else:
                b = np.random.uniform(*mr, size=X.shape[1])
                b = b.reshape(1, -1)
                b = np.repeat(b, X.shape[0], axis=0)
        # for samples we create a random normal background for each sample(row)
        # and repeat this array for each feature(column)
        else:
            if X.ndim == 1:
                b = np.random.uniform(*mr, size=1)
            else:
                b = np.random.uniform(*mr, size=X.shape[0])
                b = b.reshape(-1, 1)
                b = np.repeat(b, X.shape[1], axis=1)
        # generate a random normal background for each sample
        bkgr = np.random.normal(b, b / 10, size=X.shape)
        # add the background to the original data
        res = X + bkgr
        if X.ndim == 1:
            res = res.flatten()
        res = super().transform(res)
        return res


class PoissonBackground(BaseDataTransformer):
    """
    Adds a poisson background to the data.
    """

    def __init__(self, multiplier: float = 1):
        """
        Initializes the instance.
        The multiplier is used to scale the data before adding the poisson background.
        The scaling is important if the data is not in the form of counts.
        It is used to simulate the effect of the background on the data.
        :Parameters:

        multiplier: float
            The multiplier for the data.

        """
        BaseDataTransformer.__init__(self)
        self.multiplier = multiplier

    def fit(self, X: np.ndarray, c: np.ndarray = None) -> "PoissonBackground":
        return super().fit(X, c)

    def transform(self, X: np.ndarray) -> np.ndarray:
        multiplier = self.multiplier
        # we divide the obtained poisson background by the multiplier
        # to bring the data back to the original scale
        res = np.random.poisson(X * multiplier) / multiplier
        res = super().transform(res)
        return res


class PorodsMultiplier(BaseDataTransformer):
    """
    Multiplies the data by the Porod's law.

    The Porod's law is given by the equation:
    I(q) = A*q^(-4) + B
    where A is the Porod's constant and B is the background.
    By multiplying the data by q^4 the small intensity values are amplified.

    This multiplier is used for samples only and requires the coordinates of the data points.
    """

    def __init__(self):
        BaseDataTransformer.__init__(self)
        self.c = None

    def fit(self, X: np.ndarray, c: Optional[np.ndarray] = None) -> "PorodsNormalizer":
        if not isinstance(c, np.ndarray):
            raise TypeError("c must be a numpy array")
        if c.shape != X.shape:
            raise ValueError("c and X must have the same shape")
        self.c = c
        return super().fit(X, c)

    def transform(self, X: np.ndarray) -> np.ndarray:
        X = super().transform(X)
        c = self.c
        if c.shape != X.shape:
            raise ValueError("c and X must have the same shape")
        res = X * c**4
        res = super().transform(res)
        return res
