from dataclasses import dataclass
from typing_extensions import List, Optional, Self, Union

import ipywidgets.widgets as widgets
from IPython.display import display, Javascript

import numpy as np
from escape.utils.widgets.parameter import ParametersPanel
from escape.utils.widgets.plot import MapWidget, Plot2dWidget
from ipywidgets import AppLayout, Layout
from numpy.typing import ArrayLike
from typeguard import typechecked


def unique(
    a: ArrayLike, tol: float = 0.0, **kwargs
) -> Union[np.ndarray, tuple[np.ndarray, ...]]:
    """Find unique elements in an array with optional tolerance.

    Args:
        a: Input array
        tol: Tolerance for considering elements equal. If >0, elements within tol are considered equal
        **kwargs: Additional arguments passed to np.unique()

    Returns:
        np.ndarray or tuple: Array of unique elements, optionally with indices and counts

    Raises:
        TypeError: If tolerance is not a number
        ValueError: If tolerance is negative
    """
    if tol < 0.0:
        raise ValueError("Tolerance must be non-negative")

    if tol > 0.0:
        a = np.asarray(a)
        aint = (a * (1.0 / tol)).astype(int)
        return_index = kwargs.pop("return_index", False)
        _, idx, *out = np.unique(aint, return_index=True, **kwargs)
        unique_a = a[idx]
        if return_index:
            out = [idx, *out]
        if len(out) == 0:
            return unique_a
        return (unique_a, *out)
    return np.unique(a, **kwargs)


@dataclass(frozen=True)
class LayoutConfig:
    """Configuration options for Escape object layout.

    Attributes:
        title: Title to display in header
        max_num_cols: Maximum number of columns for multi-object layout (must be >= 1)
    """

    title: str = ""
    max_num_cols: int = 2

    def __post_init__(self):
        if self.max_num_cols < 1:
            raise ValueError("max_num_cols must be at least 1")


class _BaseLayout:
    """Base class for Escape object layouts.

    This class provides the core layout functionality and widget initialization methods.
    It is intended to be used as a mixin class with AppLayout.
    """

    center_height: int = 600
    header_height: int = 50
    footer_height: int = 100

    def init_layout(
        self,
        header_height: Optional[int] = None,
        center_height: Optional[int] = None,
        footer_height: Optional[int] = None,
    ) -> None:
        """Initialize the layout components."""
        self.pane_widths = [1, 3, 1]
        self._set_pane_heights(
            header_height=header_height,
            center_height=center_height,
            footer_height=footer_height,
        )
        self.center = self.init_center()
        self.header = self.init_header()
        self.footer = self.init_footer()
        self.left_sidebar = self.init_left_sidebar()
        self.right_sidebar = self.init_right_sidebar()

    def _set_pane_heights(
        self,
        header_height: Optional[int] = None,
        center_height: Optional[int] = None,
        footer_height: Optional[int] = None,
    ) -> None:
        """Set pane heights from configuration."""
        if header_height is None:
            ph = self.pane_heights[0]
            if isinstance(ph, str) and ph.endswith("px"):
                header_height = int(ph[:-2])
            else:
                header_height = self.header_height
        if center_height is None:
            ph = self.pane_heights[1]
            if isinstance(ph, str) and ph.endswith("px"):
                center_height = int(ph[:-2])
            else:
                center_height = self.center_height
        if footer_height is None:
            ph = self.pane_heights[2]
            if isinstance(ph, str) and ph.endswith("px"):
                footer_height = int(ph[:-2])
            else:
                footer_height = self.footer_height
        self.pane_heights = [
            f"{header_height}px",
            f"{center_height}px",
            f"{footer_height}px",
        ]

    def init_center(self) -> Optional[object]:
        """Initialize the central widget.

        Returns:
            Optional[object]: Instantiated central widget or None if no center_widget_cls defined
        """
        return None

    def init_header(self) -> Optional[widgets.HTML]:
        """Initialize the header widget.

        Returns:
            Optional[widgets.HTML]: Header widget with title or None if no title
        """
        title = self._config.title
        if title:
            return widgets.HTML(value=f"<h4>{title}</h4>")
        return None

    def init_footer(self) -> Optional[object]:
        """Initialize the footer widget.

        Returns:
            Optional[object]: Footer widget or None by default
        """
        return None

    def init_left_sidebar(self) -> Optional[object]:
        """Initialize the left sidebar widget.

        Returns:
            Optional[object]: Left sidebar widget or None by default
        """
        return None

    def init_right_sidebar(self) -> Optional[ParametersPanel]:
        """Initialize the right sidebar with parameter controls.

        Creates a ParametersPanel for the Escape object parameters if enabled.

        Returns:
            Optional[ParametersPanel]: Parameter panel or None if disabled or no parameters exist
        """
        if not self._with_parameters:
            return None

        objpanel = ParametersPanel(
            self._escobj,
            checkbox_property=self._parcheck_property,
            layout=Layout(
                overflow_y="auto",
                border="1px solid lightgrey",
                padding="5px 1px 5px 10px",
                max_height=self.pane_heights[1],
                min_height="50px",
            ),
        )

        for par in objpanel.children:
            par.slider.observe(self.update, names="value")

        return objpanel if objpanel.children else None

    def prepare_data(self) -> None:
        """Prepare data for visualization.

        This method should be overridden by subclasses to implement custom data preparation
        logic for visualization.

        Returns:
            None by default. Subclasses should return the data needed for visualization.
        """
        pass

    def update(self, *args) -> None:
        """Update the visualization.

        Updates the central widget if it exists and updates are enabled.
        """
        if self._no_update:
            return
        data = self.prepare_data()
        cw = self.center
        if cw is not None and data is not None:
            cw.update(*data)

    def update_parameter_sliders(self, silent: bool = False) -> None:
        """Update parameter slider values.

        Args:
            silent: If True, suppress widget updates during refresh
        """
        if silent:
            self._no_update = True
        right_sidebar = self.right_sidebar
        if right_sidebar is not None:
            for par in right_sidebar.children:
                par.pull_value()
        if silent:
            self._no_update = False

    def disable_parameters(self, val: bool) -> None:
        """Enable/disable parameter controls.

        Args:
            val: True to disable parameter controls, False to enable them
        """
        rsb = self.right_sidebar
        if rsb is not None:
            rsb.disabled = val

    def cleanup(self) -> None:
        """Clean up resources and remove widget observers."""
        if self.right_sidebar is not None:
            for par in self.right_sidebar.children:
                par.slider.unobserve(self.update, names="value")

    def config(self, title: str = "", max_num_cols: int = 2) -> Self:
        """Configure the layout with a title and maximum number of columns.

        Args:
            title: Title to display in header
            max_num_cols: Maximum number of columns for multi-object layout
        """
        config = LayoutConfig(title=title, max_num_cols=max_num_cols)
        self._config = config
        self.cleanup()
        self.init_layout()
        return self
    
    def info(self, msg: str) -> None:
        """Display an info message.

        Args:
            msg: The message to display
        """
        # Create a simple HTML widget to display the message
        msg_widget = widgets.HTML(
            value=f'<div style="padding: 10px; margin: 5px; border: 1px solid #ccc; border-radius: 4px;">{msg}</div>'
        )
        display(msg_widget)

    def error(self, msg: str) -> None:
        """Display an error message.

        Args:
            msg: The message to display
        """
        # Create a simple HTML widget to display the message
        msg_widget = widgets.HTML(
            value=f'<div style="padding: 10px; margin: 5px; border: 2px solid #dc3545; background-color: #f8d7da; color: #721c24; border-radius: 4px;">{msg}</div>'
        )
        display(msg_widget) 
    
    def warning(self, msg: str) -> None:
        """Display a warning message.

        Args:
            msg: The message to display
        """
        # Create a simple HTML widget to display the message
        msg_widget = widgets.HTML(
            value=f'<div style="padding: 10px; margin: 5px; border: 2px solid #ffc107; background-color: #fff3e0; color: #856404; border-radius: 4px;">{msg}</div>'
        )
        display(msg_widget)


class BaseLayout(_BaseLayout, AppLayout):
    """Base layout class for Escape objects visualization.

    Provides a configurable layout with parameter panels and visualization widgets.
    The layout includes:
    - A central widget area for visualization
    - Optional header and footer areas
    - Optional left and right sidebars for parameters and controls
    - Configurable pane dimensions via LayoutConfig
    """

    def __init__(
        self,
        escobj: object,
        config: Optional[LayoutConfig] = None,
        parcheck_property: Optional[str] = None,
        with_parameters: bool = True,
    ) -> None:
        """Initialize the layout.

        Args:
            escobj: The Escape object to visualize
            config: Widget configuration options
            parcheck_property: Property name for parameter checkboxes
            with_parameters: Whether to include parameter controls
        """
        AppLayout.__init__(self)

        if config is None:
            config = LayoutConfig()
        self._config = config
        self._no_update = False
        self._parcheck_property = parcheck_property or "fixed"
        self._with_parameters = with_parameters
        self._escobj = escobj
        self.layout = Layout(height="100%", overflow="hidden")
        self.init_layout()
        self.update()

    def _repr_mimebundle_(self, **kwargs) -> dict:
        """Generate representation for Jupyter display."""
        self.update()
        return super()._repr_mimebundle_(**kwargs)


class MultiObjectLayout(_BaseLayout, AppLayout):
    """Layout for displaying multiple Escape objects.

    This layout allows for the visualization of multiple Escape objects in a single
    widget. It provides a grid layout with a configurable maximum number of columns.
    """

    row_height: int = 300

    @typechecked
    def __init__(
        self,
        widgets: Union[list[BaseLayout], tuple[BaseLayout, ...]],
        config: Optional[LayoutConfig] = None,
    ) -> None:
        """Initialize the multi-object layout.

        Args:
            widgets: List of BaseLayout instances to display
            config: Widget configuration options
            config: Widget configuration options
        """
        AppLayout.__init__(self)
        if config is None:
            config = LayoutConfig()
        self._config = config
        self._widgets = widgets
        self.init_layout()
        self.update()

    def _set_pane_heights(
        self,
        header_height: Optional[int] = None,
        center_height: Optional[int] = None,
        footer_height: Optional[int] = None,
    ) -> None:
        """Set pane heights from configuration."""
        if center_height is None:
            center_height = self.center_height
        config = self._config
        n_objects = len(self._widgets)
        n_max_cols = min(config.max_num_cols, n_objects)
        ph = self.row_height * ((n_objects + n_max_cols - 1) // n_max_cols)
        center_height = max(ph, center_height)
        if header_height is None:
            header_height = self.header_height
        if footer_height is None:
            footer_height = self.footer_height
        self.pane_heights = [
            f"{header_height}px",
            f"{center_height}px",
            f"{footer_height}px",
        ]

    def _repr_mimebundle_(self, **kwargs) -> dict:
        """Generate representation for Jupyter display."""
        self.update()
        return super()._repr_mimebundle_(**kwargs)

    def init_center(self) -> Optional[object]:
        """Initialize the central widget.

        Creates a grid layout containing all child widgets, with a configurable
        maximum number of columns.

        Returns:
            Optional[object]: Grid layout containing child widgets
        """
        config = self._config
        n_objects = len(self._widgets)
        n_cols = min(config.max_num_cols, n_objects)
        n_rows = (n_objects + n_cols - 1) // n_cols
        grid = widgets.GridspecLayout(n_rows, n_cols)

        ph = int(self.pane_heights[1][:-2]) // n_rows
        for i, obj in enumerate(self._widgets):
            ph_ = ph
            row = i // n_cols
            col = i % n_cols
            obj._with_parameters = False
            if obj.header is not None:
                ph_ -= int(obj.pane_heights[0][:-2])
            if obj.footer is not None:
                ph_ -= int(obj.pane_heights[2][:-2])
            obj.init_layout(center_height=ph_)
            grid[row, col] = obj
        return grid

    def init_header(self) -> Optional[widgets.HTML]:
        """Initialize the header widget.

        Returns:
            Optional[widgets.HTML]: Header widget with title or None if no title
        """
        title = self._config.title
        if title:
            return widgets.HTML(value=f"<h2>{title}</h2>")
        return None

    def init_right_sidebar(self) -> Optional[ParametersPanel]:
        """Initialize the right sidebar with parameter controls.

        Creates a combined ParametersPanel for all child widgets' Escape objects.

        Returns:
            Optional[ParametersPanel]: Combined parameter panel or None if no parameters exist
        """
        escobjs = [w._escobj for w in self._widgets]
        objpanel = ParametersPanel(
            escobjs,
            checkbox_property="fixed",
            layout=Layout(
                overflow_y="auto",
                border="1px solid lightgrey",
                padding="5px 1px 5px 10px",
                max_height=self.pane_heights[1],
                min_height="50px",
            ),
        )
        for par in objpanel.children:
            par.slider.observe(self.update, names="value")
        if not objpanel.children:
            return None
        return objpanel

    def update(self, *args) -> None:
        """Update the visualization by updating all child widgets."""
        for w in self._widgets:
            w.update(*args)

    def config(self, title: str = "", max_num_cols: int = 2) -> Self:
        """Configure the layout with a title and maximum number of columns.

        Args:
            title: Title to display in header
            max_num_cols: Maximum number of columns for multi-object layout
        """
        config = LayoutConfig(title=title, max_num_cols=max_num_cols)
        self._config = config
        self.cleanup()
        self.init_layout()
        return self


ConfigType = Optional[LayoutConfig]

StackConfigType = Union[ConfigType, list[ConfigType], tuple[ConfigType]]


class BaseStackLayout(MultiObjectLayout):
    """Base layout for stacked Escape objects."""

    def __init__(
        self,
        escobjs: tuple[object],
        object_configs: Optional[StackConfigType] = None,
        config: Optional[LayoutConfig] = None,
        **kwargs,
    ) -> None:
        """Initialize the stack layout.

        Args:
            escobjs: List of Escape objects to display
            object_configs: List of configuration options for each Escape object
            config: Widget configuration options
        """
        if config is None:
            config = LayoutConfig()
        ws: list[BaseLayout] = []

        if object_configs is None:
            object_configs = [None] * len(escobjs)

        configs = None
        if object_configs is not None:
            if isinstance(object_configs, (tuple, list)) and len(object_configs) == len(
                escobjs
            ):
                configs = object_configs
            else:
                configs = [object_configs]
        else:
            configs = [None] * len(escobjs)

        for escobj, cfg in zip(escobjs, configs):
            w = escobj.show(config=cfg, with_parameters=False)
            ws.append(w)

        super().__init__(ws, config=config)


class BaseLayoutWithPlot(BaseLayout):
    """Layout for displaying data objects with error bars and masking support.

    Supports visualization of data with different dimensionalities:
    - 1D: Line plot with error bars
    - 2D: Heatmap visualization

    The layout handles data masking and error propagation appropriately for each case.

    """

    config_cls = None
    plot2d_cls = Plot2dWidget
    map_cls = MapWidget

    def init_center(self) -> Optional[object]:
        """Initialize the central widget.

        Returns:
            Plot2dWidget or MapWidget based on data dimensionality

        Raises:
            ValueError: If number of variables is not 1 or 2
        """
        num_variables: int = self._escobj.num_variables
        if num_variables == 1:
            return self.plot2d_cls(
                self, self._config, width=None, height=int(self.pane_heights[1][:-2])
            )
        elif num_variables == 2:
            return self.map_cls(
                self, self._config, width=None, height=int(self.pane_heights[1][:-2])
            )
        else:
            raise ValueError(
                f"Unexpected number of variables, expected one of [1, 2], got {num_variables}"
            )

    def config(
        self,
        title: str = "",
        xlabel: str = "",
        ylabel: str = "",
        xlog: bool = False,
        ylog: bool = False,
        grid: bool = False,
        cbmin: Optional[float] = None,
        cbmax: Optional[float] = None,
        cblog: bool = False,
        colorscale: str = "Viridis",
        fig_title: str = "",
        data_title: str = "",
        error_title: str = "",
        show_errors: bool = True,
    ) -> Self:
        """Configure the layout with a title and maximum number of columns.

        Args:
            title: Title to display in header
            xlabel: Label for x-axis
            ylabel: Label for y-axis
            zlabel: Label for z-axis/colorbar
            xscale: Scale for x-axis ('linear' or 'log')
            yscale: Scale for y-axis ('linear' or 'log')
            zscale: Scale for z-axis/colorbar ('linear' or 'log')
            colormap: Colormap name for 2D plots
            max_num_cols: Maximum number of columns for multi-object layout
        """
        config = self.config_cls(
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            xlog=xlog,
            ylog=ylog,
            grid=grid,
            cbmin=cbmin,
            cbmax=cbmax,
            cblog=cblog,
            colorscale=colorscale,
            fig_title=fig_title,
            data_title=data_title,
            error_title=error_title,
            show_errors=show_errors,
        )
        self._config = config
        self.cleanup()
        self.init_layout()
        return self
