from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Union

import numpy as np
from typeguard import typechecked

if TYPE_CHECKING:
    from escape.core.data import data_obj, datastack_obj

from escape.utils.widgets.base import (
    BaseLayoutWithPlot,
    BaseStackLayout,
    LayoutConfig,
    unique,
)
from escape.utils.widgets.plot import (
    MapPlotConfig,
    Plot2dConfig,
)


@dataclass(frozen=True)
class DataConfig(LayoutConfig, Plot2dConfig, MapPlotConfig):
    """Configuration for data object layout."""


class DataLayout(BaseLayoutWithPlot):
    """Layout for displaying data objects with error bars and masking support.

    Supports visualization of data with different dimensionalities:
    - 1D: Line plot with error bars
    - 2D: Heatmap visualization

    The layout handles data masking and error propagation appropriately for each case.

    """

    config_cls = DataConfig

    @typechecked
    def __init__(
        self, escobj: "data_obj", config: Optional[DataConfig] = None, **kwargs
    ):
        """Initialize data object layout.

        Args:
            escobj: Data object to visualize
            config: Configuration for data visualization

        Raises:
            ValueError: If number of variables is not 1 or 2
        """
        if config is None:
            config = DataConfig()
        super(DataLayout, self).__init__(escobj, config=config, **kwargs)

    def prepare_data(
        self,
    ) -> tuple[np.ma.MaskedArray, Optional[np.ma.MaskedArray], Optional[tuple]]:
        """Prepare data for visualization.

        Returns:
            tuple containing:
                - Masked array of y values
                - Optional masked array of error values
                - Optional extent tuple for 2D plots
        """
        escobj = self._escobj
        num_variables: int = escobj.num_variables
        xarr: np.ndarray = np.array(escobj.ini_coordinates)
        yarr: np.ndarray = np.array(escobj.ini_intensities)
        earr: np.ndarray = (
            np.array(escobj.ini_errors) if len(escobj.ini_errors) else None
        )
        marr: np.ndarray = np.array(escobj.mask)

        # Create masked arrays once for efficiency
        masked_yarr: np.ma.MaskedArray = np.ma.masked_array(yarr, ~marr)
        masked_earr: np.ma.MaskedArray = (
            np.ma.masked_array(earr, ~marr) if earr is not None else None
        )

        if num_variables <= 1:
            masked_xarr: np.ma.MaskedArray = np.ma.masked_array(xarr, ~marr)
            return masked_xarr, masked_yarr, masked_earr
        elif num_variables == 2:
            x0arr = xarr[0::num_variables]
            x1arr = xarr[1::num_variables]

            # Check if data is on a regular grid by comparing unique coordinates
            x0_unique = unique(x0arr, tol=1e-8)
            x1_unique = unique(x1arr, tol=1e-8)

            # Count points that should be in a regular grid
            expected_points = len(x0_unique) * len(x1_unique)
            actual_points = len(yarr)

            x0_min, x0_max = np.min(x0arr), np.max(x0arr)
            x1_min, x1_max = np.min(x1arr), np.max(x1arr)

            if expected_points != actual_points:
                # Data is irregular - need interpolation
                from scipy.interpolate import griddata

                # Estimate grid size based on point density
                grid_size = int(np.sqrt(len(x0arr)))
                x0_grid = np.linspace(x0_min, x0_max, grid_size)
                x1_grid = np.linspace(x1_min, x1_max, grid_size)
                X0, X1 = np.meshgrid(x0_grid, x1_grid)

                # Interpolate data onto regular grid
                points = np.column_stack((x0arr, x1arr))
                yarr_interp = griddata(points, yarr, (X0, X1), method="cubic")
                earr_interp = (
                    griddata(points, earr, (X0, X1), method="cubic")
                    if earr is not None
                    else None
                )
                marr_interp = griddata(points, marr, (X0, X1), method="nearest")

                # Update arrays with interpolated values
                yarr = yarr_interp
                earr = earr_interp
                marr = marr_interp
                extent = (x0_min, x0_max, x1_min, x1_max)

            else:
                # Data is already on regular grid - reshape directly
                rows = len(x1_unique)
                cols = len(x0_unique)

                yarr = yarr.reshape((rows, cols))
                earr = earr.reshape((rows, cols)) if earr is not None else None
                marr = marr.reshape((rows, cols))

                extent = (x0_min, x0_max, x1_min, x1_max)
            return (
                np.ma.masked_array(yarr, ~marr),
                np.ma.masked_array(earr, ~marr) if earr is not None else None,
                extent,
            )


class DataStackLayout(BaseStackLayout):
    """Layout for displaying a stack of data objects.

    Supports visualization of data with different dimensionalities:
    - 1D: Line plot with error bars
    - 2D: Heatmap visualization
    """

    @typechecked
    def __init__(
        self,
        data_stack: "datastack_obj",
        data_configs: Optional[Union[DataConfig, list[DataConfig]]] = None,
        config: Optional[LayoutConfig] = None,
        **kwargs
    ) -> None:
        """Initialize the data stack layout.

        Args:
            data_stack: Data stack object to visualize
            data_configs: Configuration options for each data object
            config: Layout configuration options
        """
        escobjs = [data_stack.data(i) for i in range(len(data_stack))]
        super().__init__(escobjs, object_configs=data_configs, config=config, **kwargs)
