import threading
from typing import TYPE_CHECKING, Optional, Union

import ipywidgets.widgets as widgets
from escape.utils.widgets.base import LayoutConfig
from escape.utils.widgets.model import ModelConfig, ModelStackLayout
from escape.utils.widgets.plot import Plot2dConfig, Plot2dWidget
from typeguard import typechecked

if TYPE_CHECKING:
    from escape.core.optimizer import optimizer_obj


class OptimizerObjectLayout(ModelStackLayout):
    """Layout for optimization visualization and control.

    Provides interactive controls for starting/stopping optimization,
    parameter manipulation, and visualization of optimization progress.

    Attributes:
        plot: Read-only property returning the plot configuration
    """

    center_height: int = 600
    header_height: int = 80
    footer_height: int = 150

    @typechecked
    def __init__(
        self,
        escobj: "optimizer_obj",
        model_configs: Optional[
            Union[ModelConfig, list[ModelConfig], tuple[ModelConfig, ...]]
        ] = None,
        config: Optional[LayoutConfig] = None,
        **kwargs,
    ):
        """Initialize the optimizer object layout.

        Args:
            escobj: The optimization object to visualize
            model_configs: Configuration options for model visualization
            config: Layout configuration options
            **kwargs: Additional keyword arguments passed to parent
        """
        if config is None:
            config = LayoutConfig()
        self._optimizer = escobj
        modelstack = escobj.modelstack
        self._progress_bar = None
        super().__init__(
            modelstack, model_configs=model_configs, config=config, **kwargs
        )
        
    def on_iteration(self):
        """Handle optimization iteration updates thread-safely."""
        try:
            self.update_parameter_sliders(silent=True)
            self.update(buf=True)
            self.update_status()
            # Update progress bar
            if self._progress_bar is not None:
                progress = self._optimizer.progress * 100
                if progress > 0:
                    self._progress_bar.value = progress
                    self._progress_bar.description = f"Progress: {progress:.1f}%"
        except Exception as err:
            self.error(f"Error in optimization iteration: {err}")

    def on_initialized(self):
        """Handle optimization initialization."""
        self.update_status()
        # Initialize progress bar
        if self._progress_bar is not None:
            self._progress_bar.value = 0
            self._progress_bar.description = "Progress: 0%"

    def on_finalized(self):
        """Handle optimization finalization."""
        try:
            self.update_parameter_sliders(silent=True)
            self.update(buf=True)
            self.update_status()
            # Set progress bar to 100% when finished
            if self._progress_bar is not None:
                self._progress_bar.value = 100
                self._progress_bar.description = "Progress: 100%"
                # Force update the progress bar
        except Exception as err:
            self.error(f"Error in optimization finalization: {err}")
        finally:
            escobj = self._optimizer
            escobj.initialization_method = self.on_initialized
            escobj.finalization_method = self.on_finalized
            escobj.iteration_method = self.on_iteration
            self._startbtn.disabled = False
            self._stopbtn.disabled = True
            self._resetbtn.disabled = False
            self._shakebtn.disabled = False
            self.disable_parameters(False)

        

    def update_status(self):
        """Update optimization status display."""
        try:
            status = self._optimizer.status_msg
            cost = self._optimizer.best_cost
            evals = self._optimizer.num_of_evaluations
            self._status_label.value = (
                f"Best cost: {cost:40.4g} " f"Evaluations: {evals} " f"Status: {status}"
            )
        except Exception as e:
            self.error(f"Error updating status: {e}")

    def init_header(self, config: Optional[LayoutConfig] = None):
        """Initialize header with optimization control buttons."""

        def on_start_click(btn):
            escobj = self._optimizer
            self._startbtn.disabled = True
            self._stopbtn.disabled = False
            self._shakebtn.disabled = True
            self._resetbtn.disabled = True
            self.disable_parameters(True)
            escobj.initialization_method = self.on_initialized
            escobj.finalization_method = self.on_finalized
            escobj.iteration_method = self.on_iteration
            # Initialize progress bar
            if self._progress_bar is not None:
                self._progress_bar.value = 0
                self._progress_bar.description = "Progress: 0%"
            threading.Thread(target=self._optimizer, args=(False,)).start()

        def on_stop_click(btn):
            self._optimizer.stop()

        def on_reset_click(btn):
            self._optimizer.reset_to_initial()
            self.update_parameter_sliders(silent=True)
            self.update(buf=False, cost_update=True)

        def on_shake_click(btn):
            self._optimizer.shake()
            self.update_parameter_sliders(silent=True)
            self.update(buf=False, cost_update=False)

        btnlay = widgets.Layout(width="60px")

        self._startbtn = widgets.Button(
            disabled=False,
            tooltip="Start optimization",
            description="Start",
            layout=btnlay,
        )
        self._startbtn.on_click(on_start_click)

        self._stopbtn = widgets.Button(
            disabled=True,
            tooltip="Stop optimization",
            description="Stop",
            layout=btnlay,
        )
        self._stopbtn.on_click(on_stop_click)

        self._resetbtn = widgets.Button(
            disabled=False,
            tooltip="Reset parameters values",
            description="Reset",
            layout=btnlay,
        )

        self._resetbtn.on_click(on_reset_click)

        self._shakebtn = widgets.Button(
            disabled=False,
            tooltip="Shake parameters values",
            description="Shake",
            layout=btnlay,
        )

        self._shakebtn.on_click(on_shake_click)

        toolbar = widgets.HBox(
            children=[
                self._startbtn,
                self._stopbtn,
                self._resetbtn,
                self._shakebtn,
            ]
        )

        
        self._progress_bar = widgets.FloatProgress(
            value=0,
            min=0,
            max=100,
            description='Progress: 0%',
            bar_style='info',
            style={'bar_color': '#007bff', 'description_width': 'initial'},
            layout=widgets.Layout(width='95%', height='30px')
        )

        return widgets.VBox(
            [toolbar, self._progress_bar]
        )

    def init_footer(self, config: Optional[LayoutConfig] = None):
        """Initialize footer with cost plot and status label."""
        plot = Plot2dConfig(xlabel="Iterations", ylabel="Cost")
        height = int(self.pane_heights[2][:-2])-50#50 is for status_label
        self._cost_plot = Plot2dWidget(
            self, config=plot, height=height
        )
        self._status_label = widgets.HTML("Status: ")
        return widgets.VBox([self._status_label, self._cost_plot])

    def update(self, buf=False, cost_update=True):
        """Update visualization with latest optimization state.

        Args:
            buf: Whether to buffer the update
            cost_update: Whether to update the cost plot
        """
        if cost_update:
            costs = self._optimizer.cost_history
            if len(costs):
                self._cost_plot.update(costs[0:, 0], costs[0:, 1])
        super(OptimizerObjectLayout, self).update(buf)
