from dataclasses import dataclass
from typing import Optional, Union

import ipywidgets as widgets
import numpy as np
import plotly.graph_objects as go
from numpy.ma import MaskedArray


@dataclass(frozen=True)
class BasePlotConfig:
    """Base class for all plot configurations.

    Provides common functionality for plotly-based plotting widgets.
    """

    xlog: bool = False
    """Whether to use a logarithmic scale for the x-axis"""
    ylog: bool = False
    """Whether to use a logarithmic scale for the y-axis"""
    grid: bool = False
    """Whether to display a grid on the plot"""
    xlabel: str = ""
    """Label for the x-axis"""
    ylabel: str = ""
    """Label for the y-axis"""
    fig_title: str = ""
    """Title for the plot"""


@dataclass(frozen=True)
class Plot2dConfig(BasePlotConfig):
    show_errors: bool = True
    """Whether to display error bars"""
    data_title: str = "Data"
    """Title for the data plot"""
    error_title: str = "Uncertainty"
    """Title for the error plot"""


@dataclass(frozen=True)
class MapPlotConfig(BasePlotConfig):
    cbmin: Optional[float] = None
    """Minimum value for the colorbar"""
    cbmax: Optional[float] = None
    """Maximum value for the colorbar"""
    cblog: bool = False
    """Whether to use a logarithmic scale for the colorbar"""
    colorscale: str = "Viridis"
    """Color scale for the plot"""


class BasePlotWidget(widgets.VBox):
    """Base class for all plot widgets.

    Provides common functionality for plotly-based plotting widgets.
    """

    def __init__(
        self,
        parent: object,
        config: Optional[BasePlotConfig] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
    ):
        """Initialize plot widget.

        Args:
            parent: Parent widget
            config: Plot configuration
        """
        super().__init__()  # Changed to use __init__ directly
        if config is None:
            config = BasePlotConfig()
        self._parent = parent

        self._figure = go.FigureWidget(
            layout=dict(
                width=width,
                height=height,
                margin=dict(t=10, b=0, l=30, r=30),
                xaxis_title=config.xlabel,
                yaxis_title=config.ylabel,
                showlegend=False,
                title=config.fig_title,
                autosize=True,
            )
        )
        if config.xlog:
            self._figure.update_xaxes(type="log")
        if config.ylog:
            self._figure.update_yaxes(type="log")

        if config.grid:
            self._figure.update_xaxes(showgrid=True)
            self._figure.update_yaxes(showgrid=True)

        self._config = config

        self.children = [self._figure]
        #self.layout = widgets.Layout(overflow='hidden')
        self.init_layout(config)

    def init_layout(self, config: BasePlotConfig):
        pass

    def update(self, *args, **kwargs):
        pass

    @property
    def width(self) -> int:
        return self._figure.layout.width

    @property
    def height(self) -> int:
        return self._figure.layout.height

    @width.setter
    def width(self, value: int):
        self._figure.update_layout(width=value)

    @height.setter
    def height(self, value: int):
        self._figure.update_layout(height=value)


class Plot2dWidget(BasePlotWidget):
    """Widget for 2D line plots with optional error bars."""

    def __init__(
        self,
        parent: object,
        config: Optional[Plot2dConfig] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
    ):
        super().__init__(parent, config, width, height)
        if config is None:
            config = Plot2dConfig()
        self._config = config

    def init_layout(self, config: Plot2dConfig):
        self._figure.add_trace(
            go.Scatter(
                x=[],
                y=[],
                mode="lines+markers",
                marker=dict(size=2),
                line=dict(width=2),
                error_y=dict(
                    type="data", array=[], visible=False, color="rgba(0,0,255,0.3)"
                ),
                showlegend=False,
            )
        )

    def update(
        self,
        xarr: Union[np.ndarray, MaskedArray],
        yarr: Union[np.ndarray, MaskedArray],
        earr: Optional[Union[np.ndarray, MaskedArray]] = None,
    ) -> None:
        """Update the plot with new data and optional error bars.

        Args:
            xarr: X-axis data
            yarr: Y-axis data
            earr: Optional error values for y-axis data
        """
        # Handle masked arrays by filtering out masked values
        if isinstance(xarr, np.ma.MaskedArray):
            xarr = xarr.compressed()
            yarr = yarr.compressed()
            if earr is not None:
                earr = earr.compressed()
        self._figure.data[0].x = xarr
        self._figure.data[0].y = yarr

        if earr is not None:
            self._figure.data[0].error_y.array = earr
            self._figure.data[0].error_y.visible = True
            self._figure.data[0].mode = "markers"
        else:
            self._figure.data[0].error_y.visible = False
            self._figure.data[0].mode = "lines+markers"


class MapWidget(BasePlotWidget):
    """Widget for 2D heatmap plots."""

    def __init__(
        self,
        parent: object,
        config: Optional[MapPlotConfig] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
    ):
        super().__init__(parent, config, width, height)
        if config is None:
            config = MapPlotConfig()
        self._config = config

    def init_layout(self, config: MapPlotConfig):
        super().init_layout(config)
        self._figure.add_trace(go.Heatmap(z=[[]], colorscale=config.colorscale))

    def _get_colorbar_settings(self, zmin, zmax, n=5):
        decimals = n if zmax <= 1 else 1
        return dict(
            title="",
            tickmode="array",
            tickvals=np.linspace(np.log10(zmin), np.log10(zmax), n),
            ticktext=np.round(
                10 ** np.linspace(np.log10(zmin), np.log10(zmax), n), decimals
            ),
        )

    def update(self, zarr, earr, extent):
        """Update the heatmap with data and optional errors.

        Args:
            zarr: 2D array of values to plot
            earr: Optional 2D array of uncertainties matching zarr shape
            extent: Optional tuple (xmin, xmax, ymin, ymax) for axis ranges
        """
        config = self._config
        z_data_orig = np.array(np.ma.filled(zarr, np.nan))
        cbmin = config.cbmin if config.cbmin is not None else np.min(z_data_orig)
        cbmax = config.cbmax if config.cbmax is not None else np.max(z_data_orig)
        z_data = z_data_orig

        if config.cblog:
            # Take log of data, handling zeros/negatives
            z_data = np.where(z_data_orig > 0, np.log10(z_data_orig), np.nan)

            # Handle errors in log scale if provided
            if earr is not None:
                # Store original values and errors for hover display
                self._figure.data[0].customdata = np.stack((z_data_orig, earr), axis=-1)
                self._figure.data[0].hovertemplate = (
                    "<i>Logplot</i><br>"
                    + "<b>X</b>: %{x}<br>"
                    + "<b>Y</b>: %{y}<br>"
                    + "<b>Z</b>: %{customdata[0]:.2e} ± %{customdata[1]:.2e}"
                )
            else:
                self._figure.data[0].customdata = z_data_orig
                self._figure.data[0].hovertemplate = (
                    "<i>Logplot</i><br>"
                    + "<b>X</b>: %{x}<br>"
                    + "<b>Y</b>: %{y}<br>"
                    + "<b>Z</b>: %{customdata:.2e}"
                )

            self._figure.data[0].colorbar = self._get_colorbar_settings(cbmin, cbmax)
            self._figure.data[0].zmin = np.log10(cbmin)
            self._figure.data[0].zmax = np.log10(cbmax)
        else:
            # Linear scale handling
            self._figure.data[0].zmin = cbmin
            self._figure.data[0].zmax = cbmax

            if earr is not None:
                # Store values and errors for hover display
                self._figure.data[0].customdata = earr
                self._figure.data[0].hovertemplate = (
                    "<b>X</b>: %{x}<br>"
                    + "<b>Y</b>: %{y}<br>"
                    + "<b>Z</b>: %{z:.2f} ± %{customdata:.2f}"
                )
            else:
                self._figure.data[0].hovertemplate = (
                    "<b>X</b>: %{x}<br>" + "<b>Y</b>: %{y}<br>" + "<b>Z</b>: %{z:.2f}"
                )

        self._figure.data[0].z = z_data

        if extent is not None:
            self._figure.data[0].x = np.linspace(extent[0], extent[1], zarr.shape[1])
            self._figure.data[0].y = np.linspace(extent[2], extent[3], zarr.shape[0])


class ModelPlot2dWidget(Plot2dWidget):
    """Widget for displaying model data with experimental data overlay."""

    def init_layout(self, config: Plot2dConfig):
        self._figure.set_subplots(
            rows=2, cols=1, row_heights=[0.8, 0.2], shared_xaxes=True
        )
        self._figure.add_trace(
            go.Scatter(
                x=[],
                y=[],
                mode="markers",
                name="Data",
                marker=dict(size=5, symbol="circle-open"),
                error_y=dict(type="data", array=[], visible=True),
            ),
            row=1,
            col=1,
        )
        self._figure.add_trace(
            go.Scatter(x=[], y=[], mode="lines", name="Model", line=dict(width=3)),
            row=1,
            col=1,
        )
        self._figure.add_trace(
            go.Scatter(x=[], y=[], mode="lines", name="Residuals", line=dict(width=3)),
            row=2,
            col=1,
        )

    def update(self, xarr, yarr, earr, sarr, rarr, model_legend="", data_legend=""):
        # Data with error bars in top subplot
        self._figure.data[0].x = xarr
        self._figure.data[0].y = yarr
        self._figure.data[0].error_y.array = earr if earr is not None else []
        self._figure.data[0].error_y.visible = earr is not None
        self._figure.data[0].name = data_legend

        # Model line in top subplot
        self._figure.data[1].x = xarr
        self._figure.data[1].y = sarr
        self._figure.data[1].name = model_legend

        # Residuals line in bottom subplot
        self._figure.data[2].x = xarr
        self._figure.data[2].y = rarr


class ModelMapWidget(MapWidget):
    """Widget for displaying 2D model data with experimental data overlay."""

    def init_layout(self, config: MapPlotConfig):
        """Initialize the layout with subplots for data, model and residuals."""
        self._figure.update_layout(
            grid=dict(
                rows=2,
                columns=2,
                column_widths=[0.7, 0.3],
                row_heights=[0.7, 0.3],
                column_spacing=0.05,
                row_spacing=0.05,
            )
        )

    def update(self, data_arr, model_arr, resid_arr, extent):
        """Update the plot with new data.

        Args:
            data_arr: 2D array of experimental data
            model_arr: 2D array of model predictions
            resid_arr: 2D array of residuals
            extent: Tuple of (xmin, xmax, ymin, ymax) for plot ranges
        """
        self._figure.data = []  # Clear existing traces

        # Main data and model overlay
        self._figure.add_trace(
            go.Heatmap(
                z=data_arr,
                colorscale="Viridis",
                name="Data",
                colorbar=dict(title="Data", x=0.45),
                hovertemplate="Data: %{z:.2e}<br>x: %{x:.2f}<br>y: %{y:.2f}",
            ),
            row=1,
            col=1,
        )

        self._figure.add_trace(
            go.Heatmap(
                z=model_arr,
                colorscale="Plasma",
                opacity=0.5,
                name="Model",
                colorbar=dict(title="Model", x=1.0),
                hovertemplate="Model: %{z:.2e}<br>x: %{x:.2f}<br>y: %{y:.2f}",
            ),
            row=1,
            col=1,
        )

        # Residuals heatmap
        self._figure.add_trace(
            go.Heatmap(
                z=resid_arr,
                colorscale="RdBu",
                zmid=0,
                name="Residuals",
                colorbar=dict(title="Residuals", x=1.0),
                hovertemplate="Residual: %{z:.2e}<br>x: %{x:.2f}<br>y: %{y:.2f}",
            ),
            row=2,
            col=1,
        )

        # Residuals histogram
        self._figure.add_trace(
            go.Histogram(
                x=resid_arr.flatten(),
                nbinsx=50,
                name="Residual Distribution",
                showlegend=False,
            ),
            row=2,
            col=2,
        )

        # Update axes ranges if extent provided
        if extent is not None:
            xmin, xmax, ymin, ymax = extent
            self._figure.update_xaxes(range=[xmin, xmax])
            self._figure.update_yaxes(range=[ymin, ymax])
